from __future__ import annotations

import gzip
from collections import defaultdict
from io import BytesIO

from typing_extensions import IO, Dict, Generator, List, Union
from warc.warc import WARCRecord

from scrapy_webarchive.cdxj import CdxjRecord
from scrapy_webarchive.wacz.constants import ARCHIVE_DIR
from scrapy_webarchive.wacz.storages import ZipStorageHandler
from scrapy_webarchive.warc import WARCReader


class WaczFile:
    """
    Handles looking up pages in the WACZ's index, and iterating over all pages in the index.
    Can also iterate over all entries in each WARC embedded in the archive.
    """

    def __init__(self, storage_handler: ZipStorageHandler):
        self.storage_handler = storage_handler
        self.index = self._parse_index(self._get_index())

    def _find_in_index(self, url: str) -> Union[CdxjRecord, None]:
        """Looks for the most relevant CDXJ record for a given URL in the index."""

        records = self.index.get(url, [])

        # If multiple entries are present, the last one is most likely to be relevant
        return records[-1] if records else None

    def get_warc_from_cdxj_record(self, cdxj_record: CdxjRecord) -> Union[WARCRecord, None]:
        """Retrieves a WARC record from the WACZ archive using a CDXJ record."""

        offset = int(cdxj_record.data["offset"])
        size = int(cdxj_record.data["length"])

        try:    
            file_part = self.storage_handler.fetch_file_part(
                file_name=ARCHIVE_DIR + cdxj_record.data["filename"],
                offset=offset,
                size=size,
            )
        except FileNotFoundError:
            return None

        warc_part = BytesIO(file_part)
        return WARCReader(warc_part).read_record()

    def get_warc_from_url(self, url: str) -> Union[WARCRecord, None]:
        """Retrieves a WARC record from the WACZ archive by searching for the URL in the index."""

        cdxj_record = self._find_in_index(url)
        return self.get_warc_from_cdxj_record(cdxj_record) if cdxj_record else None

    def iter_index(self) -> Generator[CdxjRecord, None, None]:
        """Iterates over all CDXJ records in the WACZ index."""

        for cdxj_records in self.index.values():
            for cdxj_record in cdxj_records:
                yield cdxj_record

    def _get_index(self) -> Union[gzip.GzipFile, IO]:
        """Opens the index file from the WACZ archive, checking for .cdxj, .cdxj.gz, .cdx. and .cdx.gz"""

        index_paths = [
            "indexes/index.cdxj",
            "indexes/index.cdxj.gz",
            "indexes/index.cdx",
            "indexes/index.cdx.gz",
        ]

        # Try opening each possible index file until we've found the index.
        for index_path in index_paths:
            try:
                fetched = self.storage_handler.fetch_file(index_path)

                if index_path.endswith(".gz"):
                    return gzip.open(BytesIO(fetched))
                else:
                    return BytesIO(fetched)

            except FileNotFoundError:
                continue

        raise FileNotFoundError("No valid index file found in WACZ file")

    def _parse_index(self, index_file: Union[gzip.GzipFile, IO]) -> Dict[str, List[CdxjRecord]]:
        cdxj_records = defaultdict(list)

        for line in index_file:
            cdxj_record = CdxjRecord.from_cdxline(line.decode(), wacz_file=self)
            cdxj_records[cdxj_record.data["url"]].append(cdxj_record)

        return cdxj_records


class MultiWaczFile:
    """Supports the same things as WaczFile, but handles multiple WACZ files underneath."""

    def __init__(self, wacz_files: List[WaczFile]) -> None:
        self.waczs = wacz_files

    def get_warc_from_cdxj_record(self, cdxj_record: CdxjRecord) -> Union[WARCRecord, None]:
        """Retrieves a WARC record from the WACZ file corresponding to the given `CdxjRecord`."""

        return cdxj_record.wacz_file.get_warc_from_cdxj_record(cdxj_record) if cdxj_record.wacz_file else None
        
    def get_warc_from_url(self, url: str) -> Union[WARCRecord, None]:
        """Searches through all WACZ files to find a WARC record that matches the provided URL."""

        for wacz in self.waczs:
            warc_record = wacz.get_warc_from_url(url)
            if warc_record:
                return warc_record

        return None

    def iter_index(self) -> Generator[CdxjRecord, None, None]:
        """
        Iterates over the index entries in all WACZ files, yielding `CdxjRecord` objects. Each yielded 
        record has its `wacz_file` attribute set to the corresponding WACZ file.
        """

        yield from (cdxj_record for wacz in self.waczs for cdxj_record in wacz.iter_index())



import struct

CD_HEADER_SIGNATURE = b"\x50\x4B\x01\x02"
EOCD_SIGNATURE = b"\x50\x4B\x05\x06"
EOCD_RECORD_SIZE = 22
ZIP64_EOCD_SIGNATURE = b"\x50\x4B\x06\x06"
ZIP64_EOCD_LOCATOR_SIGNATURE = b"\x50\x4B\x06\x07"
ZIP64_EOCD_RECORD_SIZE = 56


def find_eocd(search_range: bytes) -> int:
    """Find the End of Central Directory (EOCD) in the byte range."""

    eocd_offset = search_range.rfind(EOCD_SIGNATURE)
    
    if eocd_offset == -1:
        raise ValueError("EOCD not found in provided search range")

    return eocd_offset


def find_zip64_eocd(locator: bytes) -> int:
    """Find the ZIP64 End of Central Directory (EOCD)."""

    if locator[:4] != ZIP64_EOCD_LOCATOR_SIGNATURE:
        raise ValueError("ZIP64 EOCD Locator not found")

    return struct.unpack("<Q", locator[8:16])[0]


def parse_eocd(eocd: bytes) -> tuple:
    """Parse the standard EOCD record."""

    cd_size = struct.unpack("<I", eocd[12:16])[0]
    cd_start = struct.unpack("<I", eocd[16:20])[0]

    return cd_start, cd_size


def parse_zip64_eocd(zip64_eocd: bytes) -> tuple:
    """Parse the ZIP64 End of Central Directory (EOCD) record."""

    cd_size = struct.unpack("<Q", zip64_eocd[40:48])[0]
    cd_start = struct.unpack("<Q", zip64_eocd[48:56])[0]

    return cd_start, cd_size


def parse_central_directory(central_directory: bytes, get_file_header_length):
    """Parse the Central Directory and return a list of file metadata."""

    entries = {}
    offset = 0

    while offset < len(central_directory):
        signature = central_directory[offset:offset + 4]
        if signature != CD_HEADER_SIGNATURE:
            break

        file_name_length = struct.unpack("<H", central_directory[offset + 28:offset + 30])[0]
        extra_field_length = struct.unpack("<H", central_directory[offset + 30:offset + 32])[0]
        compressed_size = struct.unpack("<I", central_directory[offset + 20:offset + 24])[0]
        header_offset = struct.unpack("<I", central_directory[offset + 42:offset + 46])[0]

        # Check for ZIP64 extra fields for large files
        if compressed_size == 0xFFFFFFFF:
            compressed_size = read_zip64_extra_field(central_directory, offset)

        file_name = central_directory[offset + 46:offset + 46 + file_name_length].decode("utf-8")

        entries[file_name] = {
            "header_offset": header_offset,
            "file_header_length": get_file_header_length(header_offset),
            "compressed_size": compressed_size,
        }
        offset += 46 + file_name_length + extra_field_length

    return entries


def read_zip64_extra_field(central_directory: bytes, offset: int) -> int:
    """Read the ZIP64 extra field for large file sizes or offsets."""

    extra_field_start = offset + 46
    extra_field_data = central_directory[extra_field_start:]

    # Look for the ZIP64 extra field signature
    signature, size = struct.unpack("<HH", extra_field_data[:4])

    if signature == 0x0001:  # ZIP64 Extra Field
        compressed_size = struct.unpack("<Q", extra_field_data[4:12])[0]
        return compressed_size

    raise ValueError("ZIP64 extra field not found for large file size")


def is_zip64(eocd):
    """Determine if the EOCD indicates a ZIP64 format."""

    if eocd[:4] != EOCD_SIGNATURE:
        raise ValueError("EOCD signature not found")

    total_entries_on_disk = struct.unpack('<H', eocd[10:12])[0]
    total_entries = struct.unpack('<H', eocd[8:10])[0]
    central_dir_size = struct.unpack('<I', eocd[12:16])[0]
    central_dir_offset = struct.unpack('<I', eocd[16:20])[0]

    # Check for ZIP64 overflow indicators
    return any([
        total_entries_on_disk == 0xFFFF,
        total_entries == 0xFFFF,
        central_dir_size == 0xFFFFFFFF,
        central_dir_offset == 0xFFFFFFFF,
    ])



from contextlib import contextmanager

from scrapy.http.request import Request
from scrapy.settings import Settings
from scrapy.utils.test import get_crawler

from scrapy_webarchive.spidermiddlewares import WaczCrawlMiddleware

from . import get_test_data_path


class TestWaczCrawlMiddlewareWarc11:
    def setup_method(self):
        self.crawler = get_crawler()
        self.spider = self.crawler._create_spider("quotes")

    def _get_settings(self, **new_settings):
        settings = {
            "SW_WACZ_SOURCE_URI": get_test_data_path("warc_1_1", "quotes.wacz").as_uri(),
        }
        settings.update(new_settings)
        return Settings(settings)
    
    @contextmanager
    def _middleware(self, **new_settings):
        settings = self._get_settings(**new_settings)
        mw = WaczCrawlMiddleware(settings, self.crawler.stats)
        mw.spider_opened(self.spider)
        yield mw

    def test_wacz_archive_is_ignored_follow_original_behaviour(self):
        request = Request("https://quotes.toscrape.com")

        with self._middleware(SW_WACZ_CRAWL=False) as mw:
            out = list(mw.process_start_requests([request], self.spider))
            assert out == [request]

    def test_wacz_archive_iterates_all_records(self):
        with self._middleware(SW_WACZ_CRAWL=True) as mw:
            out = list(mw.process_start_requests([], self.spider))
            assert len(out) == 101

    def test_wacz_archive_filters_allowed_domains(self):
        setattr(self.spider, "allowed_domains", "quotes.toscrape.com")

        with self._middleware(SW_WACZ_CRAWL=True) as mw:
            out = list(mw.process_start_requests([], self.spider))
            assert len([request for request in out if "wacz_crawl_skip" not in request.flags]) == 61

    def test_wacz_archive_filters_archive_regex(self):
        setattr(self.spider, "archive_disallow_regexp", r"https://quotes\.toscrape\.com/page/\d+/")

        with self._middleware(SW_WACZ_CRAWL=True) as mw:
            out = list(mw.process_start_requests([], self.spider))
            assert len([request for request in out if "wacz_crawl_skip" not in request.flags]) == 9



from __future__ import annotations

import re
from urllib.parse import urlparse

from scrapy import Request, Spider, signals
from scrapy.crawler import Crawler
from scrapy.exceptions import NotConfigured
from scrapy.settings import Settings
from scrapy.statscollectors import StatsCollector
from typing_extensions import Iterable, Self, Union

from scrapy_webarchive.exceptions import WaczMiddlewareException
from scrapy_webarchive.wacz.storages import ZipStorageHandlerFactory
from scrapy_webarchive.wacz.wacz_file import MultiWaczFile, WaczFile
from scrapy_webarchive.warc import record_transformer


class BaseWaczMiddleware:
    """A base class for middlewares that require opening one or more WACZ files."""

    wacz: Union[WaczFile, MultiWaczFile]

    def __init__(self, settings: Settings, stats: StatsCollector) -> None:
        self.stats = stats
        wacz_uri = settings.get("SW_WACZ_SOURCE_URI", None)

        if not wacz_uri:
            raise NotConfigured

        self.wacz_uris = re.split(r"\s*,\s*", wacz_uri)
        self.crawl = settings.get("SW_WACZ_CRAWL", False)

    @classmethod
    def from_crawler(cls, crawler: Crawler) -> Self:
        assert crawler.stats
        o = cls(crawler.settings, crawler.stats)
        crawler.signals.connect(o.spider_opened, signals.spider_opened)
        return o

    def spider_opened(self, spider: Spider) -> None:
        """
        Handles the initialization of WACZ files when the spider is opened.

        This method is called when the Scrapy spider starts, and it attempts to open WACZ files 
        defined in the spider's configuration. It processes multiple WACZ URIs provided in `self.wacz_uris`, logs the 
        process, and collects valid WACZ files for further use.

        If only one WACZ URI is provided, it opens and assigns the file to `self.wacz` as a `WaczFile` instance. 
        If multiple URIs are provided, valid files are assigned to `self.wacz` as a `MultiWaczFile` instance.
        """

        spider.logger.info(f"[WACZDownloader] Found {len(self.wacz_uris)} WACZ URI(s) to open")
        wacz_files = []
        
        for wacz_uri in self.wacz_uris:
            spider.logger.info(f"[WACZDownloader] Opening WACZ {wacz_uri}")
            storage_handler = ZipStorageHandlerFactory.get_handler(wacz_uri, spider.settings)

            if not storage_handler.zip_exists:
                spider.logger.error(f"[WACZDownloader] Could not open WACZ {wacz_uri}")
                continue

            wacz_files.append(WaczFile(storage_handler=storage_handler))

        if wacz_files:
            spider.logger.info(
                f"[WACZDownloader] Continuing with {len(wacz_files)}/{len(self.wacz_uris)} valid WACZ files"
            )
            self.wacz = wacz_files[0] if len(wacz_files) == 1 else MultiWaczFile(wacz_files)

        # If there are not wacz_files, we raise a `WaczMiddlewareException` in the downloader/spider middleware.
        # Raising an exception here does not stop the job from running. If there are no valid WACZ files configured
        # we do not want to continue the job.

    def _is_off_site(self, url: str, spider: Spider) -> bool:
        """Check if the URL is off-site based on allowed domains."""

        return hasattr(spider, "allowed_domains") and urlparse(url).hostname not in spider.allowed_domains

    def _is_disallowed_by_spider(self, url: str, spider: Spider) -> bool:
        """Check if the URL is disallowed by the spider's archive rules."""

        return hasattr(spider, "archive_disallow_regexp") and not re.search(spider.archive_disallow_regexp, url)


class WaczCrawlMiddleware(BaseWaczMiddleware):
    """
    Scrapy WACZ crawl spider middleware to crawl from a WACZ archive.

    Replaces the default behaviour of the spider by solely iterating over the
    entries in the WACZ file.
    """

    def spider_opened(self, spider: Spider) -> None:
        if not self.crawl:
            return

        super().spider_opened(spider)

    def process_start_requests(self, start_requests: Iterable[Request], spider: Spider) -> Iterable[Request]:
        """Processes start requests and yields WACZ index entries or original requests based on the crawl setting."""
        
        # If crawl is disabled, yield the original start requests.
        if not self.crawl:
            yield from start_requests
            return
        # If the attribute has not been set, none of the WACZ could be opened.
        elif not hasattr(self, "wacz"):
            raise WaczMiddlewareException("Could not open any WACZ files, check your WACZ URIs and authentication.")

        # Iterate over entries in the WACZ index.
        for entry in self.wacz.iter_index():
            url = entry.data["url"]

            # Filter out off-site requests
            if self._is_off_site(url, spider):
                self.stats.inc_value("webarchive/crawl_skip/off_site", spider=spider)
                flags = ["wacz_start_request", "wacz_crawl_skip"]
            # Ignore disallowed pages (to avoid crawling e.g. redirects from whitelisted pages to unwanted ones)
            elif self._is_disallowed_by_spider(url, spider):
                self.stats.inc_value("webarchive/crawl_skip/disallowed", spider=spider)
                flags = ["wacz_start_request", "wacz_crawl_skip"]
            else:
                self.stats.inc_value("webarchive/start_request_count", spider=spider)
                flags = ["wacz_start_request"]

            yield record_transformer.request_for_record(
                entry,
                flags=flags,
                meta={"cdxj_record": entry},
                dont_filter=True,
            )



from __future__ import annotations

import hashlib
import logging
from datetime import datetime, timezone
from pathlib import Path
from typing import IO, Tuple
from urllib.parse import urlparse

WARC_DT_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
TIMESTAMP_DT_FORMAT = "%Y%m%d%H%M%S"
BUFF_SIZE = 1024 * 64

logger = logging.getLogger(__name__)

def get_formatted_dt_string(format: str) -> str:
    return datetime.now(timezone.utc).strftime(format)


def header_lines_to_dict(lines):
    # TODO: Only supports each header appearing once, not multiple occurences.
    headers = {}
    for line in lines:
        k, v = line.split(b":", 1)
        v = v.lstrip()
        headers[k] = v
    return headers


def get_scheme_from_uri(uri: str) -> str:
    if Path(uri).is_absolute():  # to support win32 paths like: C:\\some\dir.
        return "file"
    else:
        return urlparse(uri).scheme


def hash_stream(hash_type: str, stream: IO) -> Tuple[int, str]:
    """Hashes the stream with given hash_type hasher."""

    # At this moment the `hash_type` (or algorithm) that we pass will always be sha256 as it is hardcoded.
    # This check is implemented in case any other algorithms will be made available in the future.
    if hash_type not in hashlib.algorithms_guaranteed:
        raise ValueError(f"Unsupported hash type: {hash_type}")

    hasher = hashlib.new(hash_type)

    size = 0
    for chunk in iter(lambda: stream.read(BUFF_SIZE), b""):
        size += len(chunk)
        hasher.update(chunk)

    return size, f"{hash_type}:{hasher.hexdigest()}"



import hashlib
import io

import pytest

from scrapy_webarchive.utils import BUFF_SIZE, hash_stream


def test_hash_stream_with_empty_stream():
    # Test with an empty stream
    data = b""
    stream = io.BytesIO(data)
    size, result = hash_stream("sha256", stream)
    
    assert size == 0
    assert result == f"sha256:{hashlib.sha256(data).hexdigest()}"

def test_hash_stream_with_md5_algorithm():
    data = b"Hello world"
    expected_hash = hashlib.md5(data).hexdigest()
    
    stream = io.BytesIO(data)
    size, result = hash_stream("md5", stream)
    
    assert size == len(data)
    assert result == f"md5:{expected_hash}"

def test_hash_stream_with_sha256_algorithm():
    data = b"Hello world"
    expected_hash = hashlib.sha256(data).hexdigest()
    
    stream = io.BytesIO(data)
    size, result = hash_stream("sha256", stream)
    
    assert size == len(data)
    assert result == f"sha256:{expected_hash}"

def test_hash_stream_with_unsupported_hash_type():
    data = b"Hello world"
    stream = io.BytesIO(data)

    with pytest.raises(ValueError):
        hash_stream("unsupported_hash", stream)

def test_hash_stream_with_large_stream():
    data = b"a" * (2 * BUFF_SIZE)  # Twice the buffer size
    expected_hash = hashlib.sha256(data).hexdigest()
    
    stream = io.BytesIO(data)
    size, result = hash_stream("sha256", stream)
    
    assert size == len(data)
    assert result == f"sha256:{expected_hash}"



from __future__ import annotations

import socket
import uuid
from io import BytesIO
from urllib.parse import urlparse

from scrapy import __version__ as scrapy_version
from scrapy.http.request import Request
from scrapy.http.response import Response
from scrapy.responsetypes import ResponseTypes
from typing_extensions import List, Optional, Tuple
from warc import WARCReader as BaseWARCReader
from warc.warc import WARCRecord
from warcio.recordloader import ArcWarcRecord
from warcio.statusandheaders import StatusAndHeaders
from warcio.warcwriter import WARCWriter

from scrapy_webarchive.cdxj import CdxjRecord
from scrapy_webarchive.exceptions import WaczMiddlewareException
from scrapy_webarchive.utils import TIMESTAMP_DT_FORMAT, get_formatted_dt_string, header_lines_to_dict


def generate_warc_fname(prefix: str) -> str:
    """
    Returns new WARC filename based on recommendation in the warc-specification:
    https://iipc.github.io/warc-specifications/specifications/warc-format/warc-1.1/#annex-c-informative-warc-file-size-and-name-recommendations
    {prefix}-{timestamp}-{serial}-{crawlhost}.warc.gz
    """

    timestamp = get_formatted_dt_string(format=TIMESTAMP_DT_FORMAT)
    crawlhost = socket.gethostname().split(".")[0]
    # As of now we only generate one WARC file. Add serial in here to adhere to the warc specification.
    serial = '00000'

    return "-".join([prefix, timestamp, serial, crawlhost]) + ".warc.gz"


class WARCReader(BaseWARCReader):
    """WARC reader with compatibility for WARC version 1.0 and 1.1."""

    SUPPORTED_VERSIONS = ["1.0", "1.1"]


class WarcFileWriter:
    """Handles writing WARC files."""

    def __init__(self, collection_name: str, warc_fname: Optional[str] = None) -> None:
        self.collection_name = collection_name
        self.warc_fname = warc_fname or generate_warc_fname(prefix=collection_name)

    def write_record(
        self, 
        url: str, 
        record_type: str, 
        headers: List[Tuple[str, str]], 
        warc_headers: StatusAndHeaders, 
        content_type: str, 
        content: bytes, 
        http_line: str,
    ) -> ArcWarcRecord:
        """Write any WARC record (response or request) to a WARC file."""

        with open(self.warc_fname, "ab") as fh:
            writer = WARCWriter(fh, gzip=True)
            http_headers = StatusAndHeaders(statusline=http_line, headers=headers, is_http_request=True)
            payload = BytesIO(content)

            record = writer.create_warc_record(
                uri=url,
                record_type=record_type,
                payload=payload,
                http_headers=http_headers,
                warc_headers=warc_headers,
                warc_content_type=content_type,
            )
            writer.write_record(record)
        
        return record

    def write_response(self, response: Response, request: Request) -> ArcWarcRecord:
        """Write a WARC-Type: response record."""

        record_id = self.__record_id()
        warc_headers = StatusAndHeaders(
            "",
            [
                ("WARC-Type", "response"),
                ("WARC-Target-URI", response.url),
                ("WARC-Date", request.meta["WARC-Date"]),
                ("WARC-Record-ID", record_id),
            ],
            protocol="WARC/1.0",
        )

        http_line = f"HTTP/1.0 {str(response.status)}"

        headers = []
        for key in response.headers:
            val = response.headers[key]
            headers.append((key.decode(), val.decode()))

        record = self.write_record(
            url=response.url,
            record_type="response",
            content=response.body,
            content_type="application/http; msgtype=response",
            headers=headers,
            warc_headers=warc_headers,
            http_line=http_line,
        )
        return record

    def write_request(self, request: Request, concurrent_to: ArcWarcRecord):
        """Write a WARC-Type: request record."""

        record_id = self.__record_id()
        warc_headers = StatusAndHeaders(
            "",
            [
                ("WARC-Type", "request"),
                ("WARC-Target-URI", request.url),
                ("WARC-Date", request.meta["WARC-Date"]),
                ("WARC-Record-ID", record_id),
                ("WARC-Concurrent-To", concurrent_to.rec_headers.get_header('WARC-Record-ID')),
            ],
            protocol="WARC/1.0",
        )

        http_line = f"{request.method} {urlparse(request.url).path} HTTP/1.0"

        headers = []
        for key in request.headers:
            val = request.headers[key]
            headers.append((key.decode(), val.decode()))

        record = self.write_record(
            url=request.url,
            record_type="request",
            content_type="application/http; msgtype=request",
            content=request.body,
            headers=headers,
            warc_headers=warc_headers,
            http_line=http_line,
        )
        return record

    def write_warcinfo(self, robotstxt_obey: bool) -> None:
        """Write WARC-Type: warcinfo record."""

        content = {
            "software": f"Scrapy/{scrapy_version} (+https://scrapy.org)",
            "format": "WARC file version 1.0",
            "conformsTo": "https://iipc.github.io/warc-specifications/specifications/warc-format/warc-1.0/",
            "isPartOf": self.collection_name,
            "robots": "obey" if robotstxt_obey else "ignore",
        }

        with open(self.warc_fname, "ab") as fh:
            writer = WARCWriter(fh, gzip=True)
            record = writer.create_warcinfo_record(filename=self.warc_fname, info=content)
            writer.write_record(record)

    @staticmethod
    def __record_id() -> str:
        """Returns WARC-Record-ID (globally unique UUID) as a string."""

        return f"<urn:uuid:{uuid.uuid1()}>"


class WarcRecordTransformer:
    """A helper class to convert WARC records into Scrapy requests and responses."""
    
    response_types = ResponseTypes()

    def request_for_record(self, cdxj_record: CdxjRecord, **kwargs):
        """Create a Scrapy request instance from a WARCRecord."""

        # TODO: locate request in WACZ and include all relevant things (like headers)
        return Request(url=cdxj_record.data["url"], method=cdxj_record.data.get("method", "GET"), **kwargs)

    def response_for_record(self, warc_record: WARCRecord, **kwargs):
        """Create a Scrapy response instance from a WARCRecord."""

        # We expect a response.
        # https://iipc.github.io/warc-specifications/specifications/warc-format/warc-1.1/#warc-type-mandatory
        if warc_record.type != "response":
            raise WaczMiddlewareException(f"Unexpected record type: {warc_record.type}")

        # We only know how to handle application/http.
        # https://iipc.github.io/warc-specifications/specifications/warc-format/warc-1.1/#content-type
        record_content_type = (warc_record["Content-Type"] or "").split(";", 1)[0]
        if record_content_type != "application/http":
            raise WaczMiddlewareException(f"Unexpected record content-type: {record_content_type}")

        # There is a date field in record['WARC-Date'], but don't have a use for it now.
        # https://iipc.github.io/warc-specifications/specifications/warc-format/warc-1.1/#warc-date-mandatory

        payload = warc_record.payload.read()
        payload_parts = payload.split(b"\r\n\r\n", 1)
        header_lines = payload_parts[0] if len(payload_parts) > 0 else ""
        body = payload_parts[1] if len(payload_parts) > 1 else None

        header_lines = header_lines.split(b"\r\n")
        header_parts = header_lines[0].split(None, 2)
        protocol = header_parts[0] if len(header_parts) > 0 else None
        status = header_parts[1] if len(header_parts) > 1 else None
        headers = header_lines_to_dict(header_lines[1:])

        if not status or not protocol:
            return None

        response_cls = self.response_types.from_headers(headers)

        return response_cls(
            url=warc_record.url,
            status=int(status.decode()),
            protocol=protocol.decode(),
            headers=headers,
            body=body,
            **kwargs,
        )


record_transformer = WarcRecordTransformer()



import gzip
import socket
from io import BytesIO

import pytest
from freezegun import freeze_time
from scrapy.http import HtmlResponse
from scrapy.http.request import Request
from scrapy.http.response import Response
from warc.warc import WARCRecord
from warcio.recordloader import ArcWarcRecordLoader

from scrapy_webarchive.cdxj import CdxjRecord
from scrapy_webarchive.exceptions import WaczMiddlewareException
from scrapy_webarchive.warc import WarcFileWriter, generate_warc_fname, record_transformer


@freeze_time("2024-10-04 08:27:11")
def test_generate_warc_fname(monkeypatch):
    prefix = "example"
    monkeypatch.setattr(socket, "gethostname", lambda: "example.local")
    assert generate_warc_fname(prefix) == "example-20241004082711-00000-example.warc.gz"


@pytest.fixture
def warc_record_response():
    payload = (
        b"HTTP/1.0 200\r\nContent-Length: 11064\r\nDate: Mon, 07 Oct 2024 09:58:44 GMT\r\nContent-Type: text/html; "
        b"charset=utf-8\r\nStrict-Transport-Security: max-age=0; includeSubDomains; preload\r\n\r\n<!DOCTYPE html>\n"
        b"<html lang=\"en\">Welcome to scrapy-webarchive!</html>"
    )
    return WARCRecord(payload=payload, headers={"WARC-Target-URI": "http://example.com"})


@pytest.fixture
def warc_record_request():
    return WARCRecord(payload=b"Welcome to scrapy-webarchive!", headers={"WARC-Type": "request"}) 


class TestWarcRecordTransformer:
    def test_request_for_record(self):
        record = CdxjRecord(
            wacz_file=None,
            surt="com,example)/index",
            host="example",
            data={
            "url": "http://example.com", 
                "mime": "text/html", 
                "status": "200", 
                "digest": "sha1:AA7J5JETQ4H7GG22MU2NCAUO6LM2EPEU", 
                "length": "2302", 
                "offset": "384", 
                "filename": "example-20241007095844-00000-BA92-CKXFG4FF6H.warc.gz",
            }
        )
        
        request = record_transformer.request_for_record(record)
        assert isinstance(request, Request)
        assert request.url == "http://example.com"
        assert request.method == "GET"

    def test_response_for_record_invalid_response_type(self, warc_record_request):
        with pytest.raises(WaczMiddlewareException):
            record_transformer.response_for_record(warc_record_request)

    def test_response_for_record(self, warc_record_response):
        response = record_transformer.response_for_record(warc_record_response)
        assert isinstance(response, HtmlResponse)
        assert response.url == "http://example.com"
        assert response.status == 200
        assert response.body == b'<!DOCTYPE html>\n<html lang="en">Welcome to scrapy-webarchive!</html>'


UTF8_PAYLOAD = u'\
HTTP/1.0 200 OK\r\n\
Content-Type: text/plain; charset="UTF-8"\r\n\
Content-Disposition: attachment; filename="example.txt"\r\n\
Custom-Header: somevalue\r\n\
Unicode-Header: %F0%9F%93%81%20text%20%F0%9F%97%84%EF%B8%8F\r\n\
\r\n\
some\n\
text'

content_length = len(UTF8_PAYLOAD.encode('utf-8'))

UTF8_RECORD = u'\
WARC/1.0\r\n\
WARC-Type: response\r\n\
WARC-Record-ID: <urn:uuid:12345678-feb0-11e6-8f83-68a86d1772ce>\r\n\
WARC-Target-URI: http://example.com/\r\n\
WARC-Date: 2000-01-01T00:00:00Z\r\n\
WARC-Payload-Digest: sha1:B6QJ6BNJ3R4B23XXMRKZKHLPGJY2VE4O\r\n\
WARC-Block-Digest: sha1:KMUABC6URWIQ7QXCZDQ5FS6WIBBFRORR\r\n\
Content-Type: application/http; msgtype=response\r\n\
Content-Length: {0}\r\n\
\r\n\
{1}\r\n\
\r\n\
'.format(content_length, UTF8_PAYLOAD)


class TestWarcFileWriter:
    warc_fname = "/tmp/test.warc.gz"
    collection_name = "example"

    def setup_method(self):
        self.writer = WarcFileWriter(collection_name=self.collection_name, warc_fname=self.warc_fname)

    def test_write_warcinfo_record(self, fs):
        fs.create_file(self.warc_fname)

        # Confidence test
        warcinfo = gzip.open(self.warc_fname, "rb").read().decode()
        assert warcinfo == ""

        # Write warcinfo record and check output
        self.writer.write_warcinfo(robotstxt_obey=True)
        warcinfo = gzip.open(self.warc_fname, "rb").read().decode()
        
        assert "WARC/1.0" in warcinfo
        assert "WARC-Type: warcinfo" in warcinfo
        assert "WARC-Record-ID:" in warcinfo
        assert "WARC-Filename: /tmp/test.warc.gz" in warcinfo
        assert "Content-Type: application/warc-fields" in warcinfo
        assert f"isPartOf: {self.collection_name}" in warcinfo
        assert "robots: obey" in warcinfo

    def test_write_warcrequest_record(self, fs):
        fs.create_file(self.warc_fname)

        # Confidence test
        warc_data = gzip.open(self.warc_fname, 'rb').read().decode()
        assert warc_data == ''

        # Write request record and check output
        request = Request("http://www.example.com", meta={'WARC-Date': '2000-01-01T00:00:00Z'})
        warc_response_record = ArcWarcRecordLoader().parse_record_stream(BytesIO(UTF8_RECORD.encode('utf-8')))
        self.writer.write_request(request=request, concurrent_to=warc_response_record)
        warc_request_record = gzip.open(self.warc_fname, 'rb').read().decode()

        assert 'WARC/1.0' in warc_request_record
        assert 'WARC-Type: request' in warc_request_record
        assert 'WARC-Record-ID:' in warc_request_record
        assert 'WARC-Concurrent-To: <urn:uuid:12345678-feb0-11e6-8f83-68a86d1772ce>' in warc_request_record
        assert 'Content-Type: application/http; msgtype=request' in warc_request_record

    def test_write_warcresponse_record(self, fs):
        fs.create_file(self.warc_fname)

        # Confidence test
        warc_data = gzip.open(self.warc_fname, 'rb').read().decode()
        assert warc_data == ''

        # Write response record and check output
        request = Request("http://www.example.com", meta={'WARC-Date': '2000-01-01T00:00:00Z'})
        response = Response("http://www.example.com")
        self.writer.write_response(request=request, response=response)
        warc_response_record = gzip.open(self.warc_fname, 'rb').read().decode()

        assert 'WARC/1.0' in warc_response_record
        assert 'WARC-Type: response' in warc_response_record
        assert 'WARC-Record-ID:' in warc_response_record
        assert 'Content-Type: application/http; msgtype=response' in warc_response_record



class WaczMiddlewareException(Exception):
    """Indicates a critical issue in the middleware."""

    pass


class UnsupportedURIException(Exception):
    """Raised when the given URI scheme is not supported by the factory."""

    pass



import gzip
import os
import struct
import zipfile
from abc import ABC, abstractmethod
from io import BytesIO
from urllib.parse import unquote, urlparse

from scrapy.exceptions import NotConfigured
from scrapy.utils.boto import is_botocore_available

from scrapy_webarchive.exceptions import UnsupportedURIException
from scrapy_webarchive.wacz import zip_utils


class ZipStorageHandler(ABC):
    """Abstract Base Class for ZIP storage handlers."""

    @abstractmethod
    def fetch_file(self, file_name: str) -> bytes:
        """Fetch the entire content of a file."""

        pass

    @abstractmethod
    def fetch_file_part(self, file_name: str, offset: int, size: int) -> bytes:
        """Fetch a specific part of a file."""

        pass

    @property
    @abstractmethod
    def zip_exists(self) -> bool:
        """Check if the ZIP file exists."""

        pass


class LocalZipStorageHandler(ZipStorageHandler):
    """Handles ZIP files stored locally."""

    def __init__(self, uri: str):
        """Initialize the handler with a local ZIP file."""

        self.uri = uri
        parsed_uri = urlparse(self.uri)
        self.file_path = unquote(parsed_uri.path)

        if not self.zip_exists:
            raise FileNotFoundError(f"ZIP file not found at {uri}")

        self.zip_file = zipfile.ZipFile(self.file_path)

    def fetch_file_part(self, file_name: str, offset: int, size: int) -> bytes:
        if file_name not in self.zip_file.namelist():
            raise FileNotFoundError(f"File {file_name} not found in ZIP archive")

        with self.zip_file.open(file_name) as file_obj:
            file_obj.seek(offset)

            if file_name.endswith(".gz"):
                with gzip.open(file_obj) as gz_file:
                    return gz_file.read()

            return file_obj.read()

    def fetch_file(self, file_name: str) -> bytes:
        if file_name not in self.zip_file.namelist():
            raise FileNotFoundError(f"File {file_name} not found in ZIP archive")

        with self.zip_file.open(file_name) as file_obj:
            if file_name.endswith(".gz"):
                with gzip.open(file_obj) as gz_file:
                    return gz_file.read()
            return file_obj.read()

    @property
    def zip_exists(self) -> bool:
        return os.path.exists(self.file_path)


class RemoteZipStorageHandler(ZipStorageHandler):
    """Abstract base class for remote ZIP file handlers."""

    @abstractmethod
    def get_object(self, range_bytes: str) -> bytes:
        """Fetch an object (or part of it) from remote storage."""

        pass

    @abstractmethod
    def get_file_info(self) -> dict:
        """Get metadata about the remote file."""

        pass

    def _find_eocd(self, file_size: int) -> int:
        """Find the End of Central Directory (EOCD) in the ZIP file."""

        search_offset = max(0, file_size - 65536)
        range_bytes = f"bytes={search_offset}-{file_size - 1}"
        search_range = self.get_object(range_bytes)
        return zip_utils.find_eocd(search_range) + search_offset

    def _find_zip64_eocd(self, eocd_offset: int) -> int:
        """Find the ZIP64 End of Central Directory (EOCD)."""

        locator_offset = eocd_offset - 20
        range_bytes = f"bytes={locator_offset}-{eocd_offset - 1}"
        locator = self.get_object(range_bytes)
        return zip_utils.find_zip64_eocd(locator) + locator_offset

    def _get_file_header_length(self, offset: int) -> int:
        """Retrieve the length of a file header from the ZIP file."""

        range_bytes = f"bytes={offset}-{offset + 29}"
        local_header = self.get_object(range_bytes)

        if local_header[:4] != b"\x50\x4b\x03\x04":
            raise ValueError("Invalid Local File Header signature")

        file_name_length, extra_field_length = struct.unpack("<HH", local_header[26:30])
        return 30 + file_name_length + extra_field_length

    def _get_zip_metadata(self) -> dict:
        """Retrieve metadata from the ZIP file, including central directory information."""

        file_size = self.get_file_info()["ContentLength"]
        eocd_offset = self._find_eocd(file_size)
        range_bytes = f"bytes={eocd_offset}-{eocd_offset + zip_utils.EOCD_RECORD_SIZE - 1}"
        eocd = self.get_object(range_bytes)

        if zip_utils.is_zip64(eocd):
            zip64_eocd_offset = self._find_zip64_eocd(eocd_offset)
            range_bytes = f"bytes={zip64_eocd_offset}-{zip64_eocd_offset + zip_utils.ZIP64_EOCD_RECORD_SIZE - 1}"
            zip64_eocd = self.get_object(range_bytes)
            cd_start, cd_size = zip_utils.parse_zip64_eocd(zip64_eocd)
        else:
            cd_start, cd_size = zip_utils.parse_eocd(eocd)

        range_bytes = f"bytes={cd_start}-{cd_start + cd_size - 1}"
        central_directory = self.get_object(range_bytes)

        return zip_utils.parse_central_directory(central_directory, self._get_file_header_length)


class S3ZipStorageHandler(RemoteZipStorageHandler):
    """Handles ZIP files stored in Amazon S3."""
    
    AWS_ACCESS_KEY_ID = None
    AWS_SECRET_ACCESS_KEY = None
    AWS_SESSION_TOKEN = None
    AWS_ENDPOINT_URL = None
    AWS_REGION_NAME = None
    AWS_USE_SSL = None
    AWS_VERIFY = None

    def __init__(self, uri: str):
        if not is_botocore_available():
            raise NotConfigured("missing botocore library")

        parse_result = urlparse(uri, allow_fragments=False)
        if parse_result.scheme != "s3":
            raise ValueError(f"Incorrect URI scheme in {uri}, expected 's3'")

        self.bucket = parse_result.netloc
        self.path = parse_result.path.lstrip("/")
        self.s3_client = self._initialize_s3_client()
        self.zip_metadata = self._get_zip_metadata()

    def _initialize_s3_client(self):
        import botocore.session
        session = botocore.session.get_session()
        return session.create_client(
            "s3",
            aws_access_key_id=self.AWS_ACCESS_KEY_ID,
            aws_secret_access_key=self.AWS_SECRET_ACCESS_KEY,
            aws_session_token=self.AWS_SESSION_TOKEN,
            endpoint_url=self.AWS_ENDPOINT_URL,
            region_name=self.AWS_REGION_NAME,
            use_ssl=self.AWS_USE_SSL,
            verify=self.AWS_VERIFY,
        )

    def get_object(self, range_bytes: str) -> bytes:
        response = self.s3_client.get_object(Bucket=self.bucket, Key=self.path, Range=range_bytes)
        return response["Body"].read()

    def get_file_info(self) -> dict:
        return self.s3_client.head_object(Bucket=self.bucket, Key=self.path)

    def fetch_file_part(self, file_name: str, offset: int, size: int) -> bytes:
        metadata = self.zip_metadata[file_name]
        range_bytes = self._calculate_range(metadata, offset, size)
        data = self.get_object(range_bytes)

        return gzip.open(BytesIO(data)).read() if file_name.endswith(".gz") else data

    def fetch_file(self, file_name: str) -> bytes:
        metadata = self.zip_metadata[file_name]
        range_bytes = self._calculate_range(metadata, 0, metadata["compressed_size"])
        return self.get_object(range_bytes)

    def _calculate_range(self, metadata: dict, offset: int, size: int) -> str:
        start = metadata["header_offset"] + metadata["file_header_length"] + offset
        end = start + size - 1
        return f"bytes={start}-{end}"

    @property
    def zip_exists(self) -> bool:
        return bool(self.get_file_info())


class ZipStorageHandlerFactory:
    @staticmethod
    def get_handler(uri: str, settings) -> ZipStorageHandler:
        if uri.startswith('s3://'):
            handler = S3ZipStorageHandler
            handler.AWS_ACCESS_KEY_ID = settings["AWS_ACCESS_KEY_ID"]
            handler.AWS_SECRET_ACCESS_KEY = settings["AWS_SECRET_ACCESS_KEY"]
            handler.AWS_SESSION_TOKEN = settings["AWS_SESSION_TOKEN"]
            handler.AWS_ENDPOINT_URL = settings["AWS_ENDPOINT_URL"]
            handler.AWS_REGION_NAME = settings["AWS_REGION_NAME"]
            handler.AWS_USE_SSL = settings["AWS_USE_SSL"]
            handler.AWS_VERIFY = settings["AWS_VERIFY"]
            return handler(uri)
        elif uri.startswith('file://'):
            return LocalZipStorageHandler(uri)
        else:
            raise UnsupportedURIException(f"URI scheme not supported: {uri}")



from __future__ import annotations

import gzip
import io
import json
import os
import zipfile
from typing import Any

from scrapy import __version__ as scrapy_version
from typing_extensions import IO, TYPE_CHECKING, Dict, List, Union

from scrapy_webarchive import __version__ as scrapy_webarchive_version
from scrapy_webarchive.cdxj import write_cdxj_index
from scrapy_webarchive.utils import (
    TIMESTAMP_DT_FORMAT,
    WARC_DT_FORMAT,
    get_formatted_dt_string,
    hash_stream,
)
from scrapy_webarchive.wacz.constants import ARCHIVE_DIR, INDEXES_DIR, WACZ_VERSION
from scrapy_webarchive.warc import WARCReader

if TYPE_CHECKING:
    from scrapy_webarchive.extensions import FilesStoreProtocol


class WaczFileCreator:
    """Handles creating WACZ archives."""

    hash_type = "sha256"
    datapackage_fname = "datapackage.json"

    def __init__(
        self, 
        store: 'FilesStoreProtocol', 
        warc_fname: str, 
        collection_name: str, 
        title: str, 
        description: str, 
        wacz_fname: Union[str, None], 
        cdxj_fname: str = "index.cdxj",
    ) -> None:
        self.store = store
        self.warc_fname = warc_fname
        self.cdxj_fname = cdxj_fname
        self.collection_name = collection_name
        self._title = title
        self._description = description
        self.wacz_fname = wacz_fname or self.get_wacz_fname()

    def create(self) -> None:
        """Create the WACZ file from the WARC and CDXJ index and save it in the configured store."""

        # Write cdxj index to a temporary file
        write_cdxj_index(output=self.cdxj_fname, inputs=[self.warc_fname])

        # Create the WACZ archive in memory
        zip_buffer = self.create_wacz_zip()

        # Cleanup the temporary files
        self.cleanup_files(self.cdxj_fname, self.warc_fname)

        # Save WACZ to the storage
        zip_buffer.seek(0)
        self.store.persist_file(path=self.wacz_fname, buf=zip_buffer, info=None)

    def create_wacz_zip(self) -> io.BytesIO:
        """Create the WACZ zip file and return the in-memory buffer."""

        zip_buffer = io.BytesIO()

        with zipfile.ZipFile(zip_buffer, "w", zipfile.ZIP_STORED, allowZip64=True) as zip_file:
            self.write_to_zip(zip_file, self.cdxj_fname, INDEXES_DIR)
            self.write_to_zip(zip_file, self.warc_fname, ARCHIVE_DIR)
            self.write_datapackage(zip_file)

        return zip_buffer

    def write_to_zip(self, zip_file: zipfile.ZipFile, filename: str, destination: str) -> None:
        """Helper function to write a file into the ZIP archive."""

        with open(filename, "rb") as file_handle:
            zip_file.writestr(destination + os.path.basename(filename), file_handle.read())

    def cleanup_files(self, *files: str) -> None:
        """Remove files from the filesystem."""

        for file in files:
            os.remove(file)

    def get_wacz_fname(self) -> str:
        """Generate WACZ filename based on the WARC filename."""

        return f"{self.collection_name}-{get_formatted_dt_string(format=TIMESTAMP_DT_FORMAT)}.wacz"
    
    def write_datapackage(self, zip_file: zipfile.ZipFile) -> None:
        """Main function to create and write the datapackage.json."""

        package_dict = self.create_package_dict()

        with zip_file.open(ARCHIVE_DIR + self.warc_fname) as warc_fh:
            package_dict = self.update_package_metadata_from_warc(warc_fh, package_dict)

        package_dict["resources"] = self.collect_resources(zip_file)

        zip_file.writestr(self.datapackage_fname, json.dumps(package_dict, indent=2))

    def create_package_dict(self) -> Dict[str, Any]:
        """Creates the initial package dictionary."""

        dt_string = get_formatted_dt_string(format=WARC_DT_FORMAT)
        return {
            "profile": "data-package",
            "title": self.title,
            "description": self.description,
            "created": dt_string,
            "modified": dt_string,
            "wacz_version": WACZ_VERSION,
            "software": f"scrapy-webarchive/{scrapy_webarchive_version}, Scrapy/{scrapy_version}",
        }

    def update_package_metadata_from_warc(self, warc_fh: IO, package_dict: Dict[str, Any]) -> Dict[str, Any]:
        """Updates the package dictionary with metadata from the WARC records."""

        warc_reader = WARCReader(gzip.open(warc_fh)) if self.warc_fname.endswith(".gz") else WARCReader(warc_fh)

        while True:
            warc_record = warc_reader.read_record()
            if warc_record is None:
                break
            
            if warc_record.type == "request":
                package_dict.update({
                    "mainPageUrl": warc_record.url,
                    "mainPageDate": warc_record.date,
                })
                break

        return package_dict

    def collect_resources(self, zip_file: zipfile.ZipFile) -> List[Dict[str, Any]]:
        """Collects resource information from the zip file."""

        resources = []

        for zip_entry in zip_file.infolist():
            with zip_file.open(zip_entry, "r") as stream:
                size, hash_ = hash_stream(self.hash_type, stream)

                resources.append({
                    "name": os.path.basename(zip_entry.filename).lower(),
                    "path": zip_entry.filename,
                    "hash": hash_,
                    "bytes": size,
                })

        return resources
    
    @property
    def title(self):
        return self._title or self.collection_name

    @property
    def description(self):
        return (
            self._description
            or f"This is the web archive generated by a scrapy-webarchive extension for the {self.collection_name} "
            "spider. It is mainly for scraping purposes as it does not contain any js/css data. Though it can be "
            "replayed as bare HTML if the site does not depend on JavaScript."
        )



from contextlib import contextmanager

from scrapy.http.request import Request
from scrapy.settings import Settings
from scrapy.spiders import Spider
from scrapy.utils.test import get_crawler

from scrapy_webarchive.downloadermiddlewares import WaczMiddleware

from . import get_test_data_path


class BaseTestWaczMiddleware:
    def setup_method(self):
        self.crawler = get_crawler(Spider)
        self.spider = self.crawler._create_spider("quotes")

    def _get_wacz_source_url(self) -> str:
        """Return the WACZ source URL for this test. Override in subclasses."""

        raise NotImplementedError

    def _get_settings(self, **new_settings):
        settings = {
            "SW_WACZ_SOURCE_URI": self._get_wacz_source_url(),
            "SW_WACZ_CRAWL": True,
        }
        settings.update(new_settings)
        return Settings(settings)

    @contextmanager
    def _middleware(self, **new_settings):
        settings = self._get_settings(**new_settings)
        mw = WaczMiddleware(settings, self.crawler.stats)
        mw.spider_opened(self.spider)
        yield mw


class TestWaczMiddleware(BaseTestWaczMiddleware):
    def _get_wacz_source_url(self):
        return get_test_data_path("warc_1_1", "quotes.wacz").as_uri()

    def test_retrieve_from_wacz_record_not_found(self):
        request = Request("http://www.example.com/")
        with self._middleware() as mw:
            response = mw.process_request(request, self.spider)
            assert response
            assert response.status == 404

    def test_retrieve_from_wacz(self):
        request = Request("https://quotes.toscrape.com/tag/love/")
        with self._middleware() as mw:
            response = mw.process_request(request, self.spider)
            assert response
            assert response.status == 200


class TestWaczMiddlewareMultiWacz(BaseTestWaczMiddleware):
    def _get_wacz_source_url(self):
        wacz_1 = get_test_data_path("warc_1_1", "quotes.wacz").as_uri()
        wacz_2 = get_test_data_path("warc_1_1", "goodreads.wacz").as_uri()
        return f'{wacz_1},{wacz_2}'

    def test_retrieve_from_first_wacz(self):
        request = Request("https://quotes.toscrape.com/tag/love/")
        with self._middleware() as mw:
            response = mw.process_request(request, self.spider)
            assert response
            assert response.status == 200

    def test_retrieve_from_second_wacz(self):
        request = Request("https://www.goodreads.com/quotes")
        with self._middleware() as mw:
            response = mw.process_request(request, self.spider)
            assert response
            assert response.status == 200



from __future__ import annotations

import os
from datetime import datetime
from io import BytesIO
from typing import Tuple

from scrapy import Spider, signals
from scrapy.crawler import Crawler
from scrapy.exceptions import NotConfigured
from scrapy.http.request import Request
from scrapy.http.response import Response
from scrapy.pipelines.files import FSFilesStore, FTPFilesStore, GCSFilesStore, S3FilesStore
from scrapy.pipelines.media import MediaPipeline
from scrapy.settings import Settings
from twisted.internet.defer import Deferred
from typing_extensions import Any, Dict, Protocol, Self, Type, Union, cast

from scrapy_webarchive.utils import WARC_DT_FORMAT, get_formatted_dt_string, get_scheme_from_uri
from scrapy_webarchive.wacz.creator import WaczFileCreator
from scrapy_webarchive.warc import WarcFileWriter


class FilesStoreProtocol(Protocol):
    # This protocol will be introduced by Scrapy at a later stage. Once that happens we can drop this bit.
    # For now we implement it ourselves to improve type hints for the stores.
    def __init__(self, basedir: str): 
        ...

    def persist_file(
        self,
        path: str,
        buf: BytesIO,
        info: MediaPipeline.SpiderInfo,
        meta: Union[Dict[str, Any], None] = None,
        headers: Union[Dict[str, str], None] = None,
    ) -> Union[Deferred[Any], None]: 
        ...


class WaczExporter:
    """WACZ exporter extension that writes spider requests/responses as WARC and later compiles them to a WACZ."""

    wacz_fname = None
    STORE_SCHEMES: Dict[str, Type[FilesStoreProtocol]] = {
        "": FSFilesStore,
        "file": FSFilesStore,
        "s3": S3FilesStore,
        "gs": GCSFilesStore,
        "ftp": FTPFilesStore,
    }

    def __init__(self, settings: Settings, crawler: Crawler) -> None:
        self.settings = settings
        
        # Check configuration prerequisites
        self._check_configuration_prerequisites()

        self.stats = crawler.stats
        self.crawler = crawler
        self.spider_name = crawler.spidercls.name if hasattr(crawler.spidercls, "name") else crawler.spider.name

        # Get the store URI and configure the WACZ filename
        store_uri, self.wacz_fname  = self._retrieve_store_uri_and_wacz_fname()

        # Initialize store and writer
        self.store: FilesStoreProtocol = self._get_store(store_uri)
        self.writer = WarcFileWriter(collection_name=self.spider_name)

    def _check_configuration_prerequisites(self) -> None:
        """raises NotConfigured if essential settings or middleware configurations are incorrect."""
        
        if not self.settings.get("SW_EXPORT_URI"):
            raise NotConfigured("Missing SW_EXPORT_URI setting.")
        
        forbidden_middleware = [
            ("scrapy_webarchive.spidermiddlewares.WaczCrawlMiddleware", "SPIDER_MIDDLEWARES"),
            ("scrapy_webarchive.downloadermiddlewares.WaczMiddleware", "DOWNLOADER_MIDDLEWARES"),
        ]
        if any(middleware in self.settings.getlist(key) for middleware, key in forbidden_middleware):
            raise NotConfigured("Disable WACZ middlewares in SPIDER_MIDDLEWARES and DOWNLOADER_MIDDLEWARES.")

    def _retrieve_store_uri_and_wacz_fname(self) -> Tuple[str, Union[str, None]]:
        """Sets up the export URI based on configuration and spider context."""

        export_uri = self.settings["SW_EXPORT_URI"].format(
            spider=self.spider_name,
            **get_archive_uri_template_dt_variables(),
        )

        if os.path.isdir(export_uri):
            return export_uri, None
        else:
            export_uri, wacz_fname = os.path.split(export_uri)
            return f"{export_uri}/", wacz_fname

    def _get_store(self, store_uri: str) -> FilesStoreProtocol:
        store_cls = self.STORE_SCHEMES[get_scheme_from_uri(store_uri)]
        return store_cls(store_uri)

    @classmethod
    def from_crawler(cls, crawler: Crawler) -> Self:
        assert crawler.stats

        try:
            exporter = cls.from_settings(crawler.settings, crawler)
        except AttributeError:
            exporter = cls(crawler.settings, crawler)

        crawler.signals.connect(exporter.response_downloaded, signal=signals.response_downloaded)
        crawler.signals.connect(exporter.spider_closed, signal=signals.spider_closed)
        crawler.signals.connect(exporter.spider_opened, signal=signals.spider_opened)
        return exporter
    
    @classmethod
    def from_settings(cls, settings: Settings, crawler: Crawler):
        """
        Store configuration is based on the images/files pipeline from Scrapy. See:
        https://github.com/scrapy/scrapy/blob/d4709e41047e794c9e39968f61c5abbcddf825c4/scrapy/pipelines/images.py#L94-L116
        """

        s3store: Type[S3FilesStore] = cast(Type[S3FilesStore], cls.STORE_SCHEMES["s3"])
        s3store.AWS_ACCESS_KEY_ID = settings["AWS_ACCESS_KEY_ID"]
        s3store.AWS_SECRET_ACCESS_KEY = settings["AWS_SECRET_ACCESS_KEY"]
        s3store.AWS_SESSION_TOKEN = settings["AWS_SESSION_TOKEN"]
        s3store.AWS_ENDPOINT_URL = settings["AWS_ENDPOINT_URL"]
        s3store.AWS_REGION_NAME = settings["AWS_REGION_NAME"]
        s3store.AWS_USE_SSL = settings["AWS_USE_SSL"]
        s3store.AWS_VERIFY = settings["AWS_VERIFY"]
        s3store.POLICY = settings["FILES_STORE_S3_ACL"]

        gcs_store: Type[GCSFilesStore] = cast(Type[GCSFilesStore], cls.STORE_SCHEMES["gs"])
        gcs_store.GCS_PROJECT_ID = settings["GCS_PROJECT_ID"]
        gcs_store.POLICY = settings["FILES_STORE_GCS_ACL"] or None

        ftp_store: Type[FTPFilesStore] = cast(Type[FTPFilesStore], cls.STORE_SCHEMES["ftp"])
        ftp_store.FTP_USERNAME = settings["FTP_USER"]
        ftp_store.FTP_PASSWORD = settings["FTP_PASSWORD"]
        ftp_store.USE_ACTIVE_MODE = settings.getbool("FEED_STORAGE_FTP_ACTIVE")

        return cls(settings=settings, crawler=crawler)

    def spider_opened(self) -> None:
        self.writer.write_warcinfo(robotstxt_obey=self.settings["ROBOTSTXT_OBEY"])

    def response_downloaded(self, response: Response, request: Request, spider: Spider) -> None:
        request.meta["WARC-Date"] = get_formatted_dt_string(format=WARC_DT_FORMAT)

        # Write response WARC record
        record = self.writer.write_response(response, request)
        self.stats.inc_value("webarchive/exporter/response_written", spider=spider)
        self.stats.inc_value(
            f"webarchive/exporter/writer_status_count/{record.http_headers.get_statuscode()}", 
            spider=spider,
        )

        # Write request WARC record
        self.writer.write_request(request, concurrent_to=record)
        self.stats.inc_value("webarchive/exporter/request_written", spider=spider)

    def spider_closed(self, spider: Spider) -> None:
        wacz_creator = WaczFileCreator(
            store=self.store,
            warc_fname=self.writer.warc_fname,
            wacz_fname=self.wacz_fname,
            collection_name=spider.name,
            title=self.settings["SW_WACZ_TITLE"],
            description=self.settings["SW_WACZ_DESCRIPTION"],
        )
        wacz_creator.create()


def get_archive_uri_template_dt_variables() -> dict:
    current_date = datetime.now()

    return {
        "year": current_date.strftime("%Y"),
        "month": current_date.strftime("%m"),
        "day": current_date.strftime("%d"),
        "timestamp": current_date.strftime("%Y%m%d%H%M%S"),
    }



from __future__ import annotations

from scrapy.exceptions import IgnoreRequest
from scrapy.http.request import Request
from scrapy.http.response import Response
from scrapy.spiders import Spider

from scrapy_webarchive.exceptions import WaczMiddlewareException
from scrapy_webarchive.spidermiddlewares import BaseWaczMiddleware
from scrapy_webarchive.warc import record_transformer


class WaczMiddleware(BaseWaczMiddleware):
    """
    Scrapy downloader middleware to crawl from a WACZ archive.

    Loads the index fully into memory, but lazily loads pages.
    This helps to work with large archives, including remote ones.
    """

    def _check_ignore_conditions(self, request: Request, spider: Spider) -> None:
        """Check conditions that would lead to ignoring the request and raise IgnoreRequest if necessary."""

        # Ignore when crawling and flag indicates this request needs to be skipped during WACZ crawl.
        if "wacz_crawl_skip" in request.flags:
            self.stats.inc_value("webarchive/crawl_skip", spider=spider)
            raise IgnoreRequest()

        # Filter out off-site requests.
        if self._is_off_site(request.url, spider):
            self.stats.inc_value("webarchive/crawl_skip/off_site", spider=spider)
            raise IgnoreRequest()

        # Ignore disallowed pages (to avoid crawling e.g. redirects from whitelisted pages to unwanted ones).
        if self._is_disallowed_by_spider(request.url, spider):
            self.stats.inc_value("webarchive/crawl_skip/disallowed", spider=spider)
            raise IgnoreRequest()

    def process_request(self, request: Request, spider: Spider):
        """Called for each request that goes through the downloader."""

        # If the attribute has not been set, none of the WACZ could be opened.
        if not hasattr(self, "wacz"):
            raise WaczMiddlewareException("Could not open any WACZ files, check your WACZ URIs and authentication.")

        # Check if the request should be ignored.
        self._check_ignore_conditions(request, spider)

        # Get record from existing index entry, or else lookup by URL.
        if request.meta.get("cdxj_record"):
            warc_record = self.wacz.get_warc_from_cdxj_record(cdxj_record=request.meta["cdxj_record"])
        else:
            warc_record = self.wacz.get_warc_from_url(url=request.url)

        # When page not found in archive, return status 404 and record it in a statistic.
        if not warc_record:
            self.stats.inc_value("webarchive/response_not_found", spider=spider)
            return Response(url=request.url, status=404)
        
        # Record found, try to re-create a response from it.
        response = record_transformer.response_for_record(warc_record)

        if not response:
            self.stats.inc_value("webarchive/response_not_recognized", spider=spider)
            raise IgnoreRequest()

        self.stats.inc_value("webarchive/hit", spider=spider)
        return response            
